from __future__ import annotations

import uuid

from datetime import date
from decimal import Decimal
from typing import TYPE_CHECKING

from sqlalchemy import Date, ForeignKey, Numeric, Uuid
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..db.base_class import reg


if TYPE_CHECKING:
    from .inventory import Inventory
    from .stock_keeping_unit import StockKeepingUnit


@reg.mapped_as_dataclass(unsafe_hash=True)
class Batch:
    __tablename__ = "batches"

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    name: Mapped[date] = mapped_column(Date, nullable=False)
    sku_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("stock_keeping_units.id"), nullable=False)
    quantity_remaining: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=2), nullable=False)
    rate: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=2), nullable=False)
    tax: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=5), nullable=False)
    discount: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=5), nullable=False)

    inventories: Mapped[list[Inventory]] = relationship("Inventory", back_populates="batch")
    sku: Mapped[StockKeepingUnit] = relationship("StockKeepingUnit", back_populates="batches")

    def __init__(
        self,
        name: date,
        quantity_remaining: Decimal,
        rate: Decimal,
        tax: Decimal,
        discount: Decimal,
        sku_id: uuid.UUID | None = None,
        sku: StockKeepingUnit | None = None,
    ):
        self.name = name
        if sku_id is not None and sku is None:
            self.sku_id = sku_id
        self.quantity_remaining = quantity_remaining
        self.rate = rate
        self.tax = tax
        self.discount = discount
        if sku is not None:
            self.sku_id = sku.id
            self.sku = sku

    def amount(self) -> Decimal:
        return self.quantity_remaining * self.rate * (1 + self.tax) * (1 - self.discount)

    @classmethod
    def suspense(cls) -> uuid.UUID:
        return uuid.UUID("a955790e-93cf-493c-a816-c7d92b127383")
