import uuid

from decimal import Decimal
from typing import TYPE_CHECKING

from sqlalchemy import ColumnElement, ForeignKey, Integer, Numeric, Uuid, type_coerce
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..db.base_class import reg


if TYPE_CHECKING:
    from .account import AccountBase
    from .cost_centre import CostCentre
    from .employee_benefit import EmployeeBenefit
    from .incentive import Incentive
    from .voucher import Voucher


@reg.mapped_as_dataclass(unsafe_hash=True)
class Journal:
    __tablename__ = "journals"

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    debit: Mapped[int] = mapped_column(Integer, nullable=False)
    amount: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=2), nullable=False)
    voucher_id: Mapped[uuid.UUID] = mapped_column(
        Uuid,
        ForeignKey("vouchers.id"),
        nullable=False,
        index=True,
    )
    account_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("accounts.id"), nullable=False)
    cost_centre_id: Mapped[uuid.UUID] = mapped_column(
        Uuid,
        ForeignKey("cost_centres.id"),
        nullable=False,
    )

    voucher: Mapped["Voucher"] = relationship("Voucher", back_populates="journals")
    account: Mapped["AccountBase"] = relationship("AccountBase", back_populates="journals")
    cost_centre: Mapped["CostCentre"] = relationship("CostCentre", back_populates="journals")
    employee_benefit: Mapped["EmployeeBenefit"] = relationship("EmployeeBenefit", back_populates="journal")
    incentive: Mapped["Incentive"] = relationship("Incentive", back_populates="journal")

    @hybrid_property
    def signed_amount(self) -> Decimal:
        return self.debit * self.amount

    @signed_amount.inplace.expression
    @classmethod
    def _signed_amount_expression(cls) -> ColumnElement[Decimal]:
        return type_coerce(cls.debit * cls.amount, Numeric(precision=15, scale=2)).label("signed_amount")

    def __init__(
        self,
        debit: int,
        amount: Decimal,
        account_id: uuid.UUID,
        cost_centre_id: uuid.UUID,
        voucher_id: uuid.UUID | None = None,
        id_: uuid.UUID | None = None,
    ) -> None:
        if id_ is not None:
            self.id = id_
        self.debit = debit
        self.amount = amount
        if voucher_id is not None:
            self.voucher_id = voucher_id
        self.account_id = account_id
        self.cost_centre_id = cost_centre_id
