from __future__ import annotations

import uuid

from decimal import Decimal
from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey, Numeric, Unicode, UniqueConstraint, Uuid
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..db.base_class import reg


if TYPE_CHECKING:
    from .batch import Batch
    from .product import Product
    from .recipe import Recipe


@reg.mapped_as_dataclass(unsafe_hash=True)
class StockKeepingUnit:
    __tablename__ = "stock_keeping_units"
    __table_args__ = (UniqueConstraint("product_id", "units"),)

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    product_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("products.id"), nullable=False)
    units: Mapped[str] = mapped_column(Unicode, nullable=False)
    fraction: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=5), nullable=False)
    product_yield: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=5), nullable=False)
    cost_price: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=2), nullable=False)
    sale_price: Mapped[Decimal] = mapped_column(Numeric(precision=15, scale=2), nullable=False)

    product: Mapped[Product] = relationship("Product", back_populates="skus")
    batches: Mapped[list[Batch]] = relationship("Batch", back_populates="sku")
    recipes: Mapped[list[Recipe]] = relationship("Recipe", back_populates="sku")

    def __init__(
        self,
        units: str,
        fraction: Decimal,
        product_yield: Decimal,
        cost_price: Decimal,
        sale_price: Decimal,
        product_id: uuid.UUID | None = None,
        product: Product | None = None,
        id_: uuid.UUID | None = None,
    ) -> None:
        if product_id is not None:
            self.product_id = product_id
        self.units = units
        self.fraction = fraction
        self.product_yield = product_yield
        self.cost_price = cost_price
        self.sale_price = sale_price
        if id_ is not None:
            self.id = id_
        if product is not None:
            self.product = product
