from collections.abc import Generator
from datetime import date, datetime, timedelta

from fastapi import APIRouter, HTTPException, Request, Security, status
from sqlalchemy import or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

import brewman.schemas.attendance as schemas

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, set_date
from ..db.session import SessionFuture
from ..models.attendance import Attendance
from ..models.employee import Employee
from ..routers.fingerprint import get_prints
from ..schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Attendance)
def attendance_blank(
    request: Request, user: UserToken = Security(get_user, scopes=["attendance"])
) -> schemas.Attendance:
    return schemas.Attendance(date_=get_date(request.session), body=[])  # type: ignore[arg-type]


@router.get("/{date_}", response_model=schemas.Attendance)
def attendance_date(
    date_: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["attendance"]),
) -> schemas.Attendance:
    set_date(date_, request.session)
    with SessionFuture() as db:
        return schemas.Attendance(
            date_=date_,  # type: ignore[arg-type]
            body=attendance_date_report(datetime.strptime(date_, "%d-%b-%Y"), db),
        )


def attendance_date_report(date_: date, db: Session) -> list[schemas.AttendanceItem]:
    body: list[schemas.AttendanceItem] = []
    employees = (
        db.execute(
            select(Employee)
            .where(
                Employee.joining_date <= date_,
                or_(
                    Employee.is_active,
                    Employee.leaving_date >= date_,
                ),
            )
            .order_by(Employee.cost_centre_id, Employee.designation, Employee.name)
        )
        .scalars()
        .all()
    )

    for item in employees:
        att = (
            db.execute(
                select(Attendance).where(
                    Attendance.employee_id == item.id,
                    Attendance.date_ == date_,
                    Attendance.is_valid == True,  # noqa: E712
                )
            )
            .scalars()
            .one_or_none()
        )

        at_type = 0 if att is None else att.attendance_type

        prints, hours_worked, full_day = get_prints(item.id, date_, db)
        body.append(
            schemas.AttendanceItem(
                id_=item.id,
                code=item.code,
                name=item.name,
                designation=item.designation,
                department=item.cost_centre.name,
                attendance_type=schemas.AttendanceType(id_=at_type),
                prints=prints,
                hours_worked=hours_worked,
                full_day=full_day,
            )
        )
    return body


@router.post("/{date_}", response_model=schemas.Attendance)
def save(
    date_: str,
    data: schemas.Attendance,
    user: UserToken = Security(get_user, scopes=["attendance"]),
) -> schemas.Attendance:
    try:
        att_date = datetime.strptime(date_, "%d-%b-%Y").date()
        with SessionFuture() as db:
            for item in data.body:
                if item.attendance_type.id_ != 0:
                    attendance = Attendance(
                        employee_id=item.id_,
                        date_=att_date,
                        attendance_type=item.attendance_type.id_,
                        user_id=user.id_,
                    )
                    attendance.create(db)
            db.commit()
            return schemas.Attendance(date_=date_, body=attendance_date_report(att_date, db))  # type: ignore[arg-type]
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def date_range(
    start: date, stop: date, step: timedelta = timedelta(days=1), inclusive: bool = False
) -> Generator[date, None, None]:
    # inclusive=False to behave like range by default
    if step.days > 0:
        while start < stop:
            yield start
            start = start + step
            # not +=! don't modify object passed in if it's mutable
            # since this function is not restricted to
            # only types from datetime module
    elif step.days < 0:
        while start > stop:
            yield start
            start = start + step
    if inclusive and start == stop:
        yield start
