import uuid

from decimal import Decimal

from fastapi import APIRouter, Security
from sqlalchemy import distinct, func, select, update
from sqlalchemy.orm import Session, contains_eager

import brewman.schemas.batch_integrity as schemas

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.batch import Batch
from ..models.cost_centre import CostCentre
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.product import Product
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.voucher import Voucher
from ..schemas.user import UserToken
from .issue import refresh_voucher


router = APIRouter()


@router.post("", response_model=list[schemas.BatchIntegrity])
def post_check_batch_integrity(
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> list[schemas.BatchIntegrity]:
    with SessionFuture() as db:
        info = negative_batches(db) + batch_dates(db)
        fix_batch_prices(db)
        db.commit()
        return info


def negative_batches(db: Session) -> list[schemas.BatchIntegrity]:
    inv_sum = func.sum(Inventory.quantity * Journal.debit).label("quantity")
    list_ = db.execute(
        select(Batch, Product.name, inv_sum)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(Batch, Product.name)  # type: ignore
        .having(Batch.quantity_remaining != inv_sum)
    ).all()

    issue = []
    for batch, product, quantity in list_:
        if quantity >= 0:
            db.execute(update(Batch).where(Batch.id == batch.id).values(quantity_remaining=quantity))
        else:
            issue.append(
                schemas.BatchIntegrity(
                    id_=batch.id,
                    product=product,
                    date_=batch.name,
                    showing=batch.quantity_remaining,
                    actual=quantity,
                    price=batch.rate,
                    details=batch_details(batch.id, db),
                )
            )
    return issue


def batch_details(batch_id: uuid.UUID, db: Session) -> list[schemas.BatchIntegrityItem]:
    list_ = db.execute(
        select(Voucher.id, Voucher.date_, Voucher.voucher_type, Inventory.quantity, Inventory.rate)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Inventory.batch_id == batch_id,
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
    ).all()
    return [
        schemas.BatchIntegrityItem(
            date_=date_,
            type_=type_.name.replace("_", " ").title(),
            url=["/", type_.name.replace("_", "-").lower(), str(id_)],
            quantity=quantity,
            price=rate,
        )
        for id_, date_, type_, quantity, rate in list_
    ]


def batch_dates(db: Session) -> list[schemas.BatchIntegrity]:
    list_ = (
        db.execute(
            select(Batch)
            .join(Batch.sku)
            .join(StockKeepingUnit.product)
            .join(Batch.inventories)
            .join(Inventory.voucher)
            .where(Voucher.date_ < Batch.name)
            .options(
                contains_eager(Batch.sku).contains_eager(StockKeepingUnit.product),
                contains_eager(Batch.inventories).contains_eager(Inventory.voucher),
            )
        )
        .unique()
        .scalars()
        .all()
    )

    issue = []
    for batch in list_:
        issue.append(
            schemas.BatchIntegrity(
                id_=batch.id,
                product=batch.sku.product.name,
                date_=batch.name,
                showing=batch.quantity_remaining,
                actual=Decimal(0),
                price=batch.rate,
                details=[
                    schemas.BatchIntegrityItem(
                        date_=inv.voucher.date_,
                        type_=inv.voucher.voucher_type.name.replace("_", " ").title(),
                        quantity=inv.quantity,
                        price=inv.rate,
                        url=[
                            "/",
                            inv.voucher.voucher_type.name.replace("_", "-").lower(),
                            str(inv.voucher.id),
                        ],
                    )
                    for inv in batch.inventories
                ],
            )
        )
    return issue


def fix_batch_prices(db: Session) -> None:
    list_ = (
        db.execute(
            select(Batch)
            .join(Batch.inventories)
            .where(Batch.rate != Inventory.rate)
            .options(contains_eager(Batch.inventories))
        )
        .unique()
        .scalars()
        .all()
    )

    for batch in list_:
        for inv in batch.inventories:
            refresh_voucher(inv.voucher_id, inv.batch_id, db)


def fix_single_batch_prices(batch_id: uuid.UUID, db: Session) -> None:
    list_: list[uuid.UUID] = list(
        db.execute(
            select(distinct(Inventory.voucher_id))
            .join(Inventory.batch)
            .where(Inventory.batch_id == batch_id, Batch.rate != Inventory.rate)
        )
        .scalars()
        .all()
    )

    for voucher_id in list_:
        refresh_voucher(voucher_id, batch_id, db)
