import uuid

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy import delete, select, update
from sqlalchemy.exc import SQLAlchemyError

import brewman.schemas.cost_centre as schemas

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.cost_centre import CostCentre
from ..models.journal import Journal
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=schemas.CostCentre)
def save(
    data: schemas.CostCentreIn,
    user: UserToken = Security(get_user, scopes=["cost-centres"]),
) -> schemas.CostCentre:
    try:
        with SessionFuture() as db:
            item = CostCentre(name=data.name)
            db.add(item)
            db.commit()
            return cost_centre_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.CostCentre)
def update_route(
    id_: uuid.UUID,
    data: schemas.CostCentreIn,
    user: UserToken = Security(get_user, scopes=["cost-centres"]),
) -> schemas.CostCentre:
    try:
        with SessionFuture() as db:
            item = db.execute(select(CostCentre).where(CostCentre.id == id_)).scalar_one()
            if item.is_fixture:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail=f"{item.name} is a fixture and cannot be edited or deleted.",
                )
            item.name = data.name
            db.commit()
            return cost_centre_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.delete("/{id_}", response_model=None)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["cost-centres"]),
) -> None:
    with SessionFuture() as db:
        item: CostCentre = db.execute(select(CostCentre).where(CostCentre.id == id_)).scalar_one()
        if item.is_fixture:
            raise HTTPException(
                status_code=status.HTTP_423_LOCKED,
                detail=f"{item.name} is a fixture and cannot be edited or deleted.",
            )
        db.execute(
            update(AccountBase)
            .where(AccountBase.cost_centre_id == item.id)
            .values(cost_centre_id=CostCentre.cost_centre_overall())
        )
        db.execute(
            update(Journal)
            .where(Journal.cost_centre_id == item.id)
            .values(cost_centre_id=CostCentre.cost_centre_overall())
        )
        db.execute(delete(CostCentre).where(CostCentre.id == item.id))
        db.commit()


@router.get("", response_model=schemas.CostCentreBlank)
def show_blank(
    user: UserToken = Security(get_user, scopes=["cost-centres"]),
) -> schemas.CostCentreBlank:
    return cost_centre_blank()


@router.get("/list", response_model=list[schemas.CostCentre])
async def show_list(user: UserToken = Depends(get_user)) -> list[schemas.CostCentre]:
    with SessionFuture() as db:
        return [
            cost_centre_info(item) for item in db.execute(select(CostCentre).order_by(CostCentre.name)).scalars().all()
        ]


@router.get("/{id_}", response_model=schemas.CostCentre)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["cost-centres"]),
) -> schemas.CostCentre:
    with SessionFuture() as db:
        item = db.execute(select(CostCentre).where(CostCentre.id == id_)).scalar_one()
        return cost_centre_info(item)


def cost_centre_info(item: CostCentre) -> schemas.CostCentre:
    return schemas.CostCentre(
        id_=item.id,
        name=item.name,
        is_fixture=item.is_fixture,
    )


def cost_centre_blank() -> schemas.CostCentreBlank:
    return schemas.CostCentreBlank(
        name="",
        is_fixture=False,
    )
