import uuid

from datetime import date, datetime
from decimal import Decimal
from math import ceil

from fastapi import APIRouter, Depends, File, HTTPException, Request, Security, status
from sqlalchemy import distinct, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

import brewman.schemas.input as schema_in
import brewman.schemas.voucher as output

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, get_last_day, set_date
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.employee import Employee
from ..models.employee_benefit import EmployeeBenefit
from ..models.journal import Journal
from ..models.validations import check_journals_are_valid
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.employee_benefit import EmployeeBenefit as EmployeeBenefitSchema
from ..schemas.user import UserToken
from . import get_lock_info
from .db_image import save_files, update_files
from .voucher import blank_voucher, check_voucher_edit_allowed, voucher_info


router = APIRouter()


@router.post("", response_model=output.Voucher)
def save_route(
    request: Request,
    data: schema_in.EmployeeBenefitIn = Depends(schema_in.EmployeeBenefitIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["employee-benefit"]),
) -> output.Voucher:
    try:
        dt = get_last_day(data.date_)
        days_in_month = dt.day
        with SessionFuture() as db:
            item: Voucher = save(data, dt, user, db)
            exp, total = save_employee_benefits(item, data.employee_benefits, days_in_month, db)
            save_journals(item, exp, total, db)
            check_journals_are_valid(item)
            save_files(item.id, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            info = voucher_info(item, db)
            return info
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(data: schema_in.EmployeeBenefitIn, date_: date, user: UserToken, db: Session) -> Voucher:
    account_types: list[int] = list(
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                AccountBase.id.in_([dj.employee.id_ for dj in data.employee_benefits])
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([data.date_], VoucherType[data.type_.replace(" ", "_").upper()], account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date_=date_,
        narration=data.narration,
        is_starred=data.is_starred,
        user_id=user.id_,
        voucher_type=VoucherType[data.type_.replace(" ", "_").upper()],
    )
    db.add(voucher)
    return voucher


def save_employee_benefits(
    voucher: Voucher,
    employee_benefits: list[EmployeeBenefitSchema],
    days_in_month: int,
    db: Session,
) -> tuple[int, int]:
    total_exp, total_total = 0, 0
    for item in employee_benefits:
        account = db.execute(select(Employee).where(Employee.id == item.employee.id_)).scalar_one()
        gross_salary = item.gross_salary
        days_worked = item.days_worked
        esi_ee, esi_er, esi_both = esi_contribution(gross_salary, days_worked, days_in_month)
        pf_ee, pf_er, pf_both = pf_contribution(gross_salary, days_worked, days_in_month)
        journal = Journal(
            amount=Decimal(esi_ee + pf_ee),
            debit=1,
            account_id=account.id,
            cost_centre_id=account.cost_centre_id,
        )
        sd = EmployeeBenefit(
            journal=journal,
            gross_salary=gross_salary,
            days_worked=days_worked,
            esi_ee=esi_ee,
            pf_ee=pf_ee,
            esi_er=esi_er,
            pf_er=pf_er,
        )
        voucher.journals.append(journal)
        voucher.employee_benefits.append(sd)
        db.add(journal)
        db.add(sd)
        total_exp += esi_er + pf_er
        total_total += esi_both + pf_both
    return total_exp, total_total


def save_journals(voucher: Voucher, exp: int, total: int, db: Session) -> None:
    account = db.execute(select(AccountBase).where(AccountBase.id == AccountBase.esi_pf_expense())).scalar_one()
    journal = Journal(
        amount=Decimal(exp),
        debit=1,
        account_id=account.id,
        cost_centre_id=account.cost_centre_id,
    )
    db.add(journal)
    voucher.journals.append(journal)
    account = db.execute(select(AccountBase).where(AccountBase.id == AccountBase.esi_pf_payable())).scalar_one()
    journal = Journal(
        amount=Decimal(total),
        debit=-1,
        account_id=account.id,
        cost_centre_id=account.cost_centre_id,
    )
    db.add(journal)
    voucher.journals.append(journal)


@router.put("/{id_}", response_model=output.Voucher)
def update_route(
    id_: uuid.UUID,
    request: Request,
    data: schema_in.EmployeeBenefitIn = Depends(schema_in.EmployeeBenefitIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["employee-benefit"]),
) -> output.Voucher:
    try:
        dt = get_last_day(data.date_)
        days_in_month = dt.day
        with SessionFuture() as db:
            item: Voucher = update_voucher(id_, data, user, db)
            if dt != item.date_:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail="Date Cannot be changed for Employee Benefit voucher!",
                )
            exp, total = update_employee_benefits(item, data.employee_benefits, days_in_month, db)
            update_journals(item, Decimal(exp), Decimal(total))
            check_journals_are_valid(item)
            update_files(item.id, data.files, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def update_voucher(id_: uuid.UUID, data: schema_in.EmployeeBenefitIn, user: UserToken, db: Session) -> Voucher:
    voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
    account_types: list[int] = list(
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                AccountBase.id.in_(
                    [dj.employee.id_ for dj in data.employee_benefits] + [vj.account_id for vj in voucher.journals]
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([voucher.date_, data.date_], voucher.voucher_type, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    check_voucher_edit_allowed(voucher, user)
    voucher.is_starred = data.is_starred
    voucher.narration = data.narration
    voucher.user_id = user.id_
    voucher.posted = False
    voucher.last_edit_date = datetime.utcnow()
    return voucher


def update_employee_benefits(
    voucher: Voucher,
    employee_benefits: list[EmployeeBenefitSchema],
    days_in_month: int,
    db: Session,
) -> tuple[int, int]:
    exp, total = 0, 0
    for i in range(len(voucher.employee_benefits), 0, -1):
        item = voucher.employee_benefits[i - 1]
        index = next((idx for (idx, d) in enumerate(employee_benefits) if d.id_ == item.id), None)
        if index is not None:
            employee_benefits.pop(index)
            exp += item.esi_er + item.pf_er
            total += item.esi_ee + item.pf_ee + item.esi_er + item.pf_er
        else:
            voucher.employee_benefits.remove(item)
            voucher.journals.remove(item.journal)
    new_exp, new_total = save_employee_benefits(voucher, employee_benefits, days_in_month, db)
    return exp + new_exp, total + new_total


def update_journals(voucher: Voucher, exp: Decimal, total: Decimal) -> None:
    journal = next(i for i in voucher.journals if i.account_id == AccountBase.esi_pf_expense())
    journal.amount = exp
    journal = next(i for i in voucher.journals if i.account_id == AccountBase.esi_pf_payable())
    journal.amount = total


@router.get("/{id_}", response_model=output.Voucher)
def get_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["employee-benefit"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=output.Voucher)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["employee-benefit"]),
) -> output.Voucher:
    additional_info = BlankVoucherInfo(date_=get_date(request.session), type_=VoucherType.EMPLOYEE_BENEFIT)  # type: ignore[arg-type]
    with SessionFuture() as db:
        return blank_voucher(additional_info, db)


def esi_contribution(gross_salary: int, days_worked: int, days_in_month: int) -> tuple[int, int, int]:
    limit = 21000
    employee_rate = 0.0175
    employer_rate = 0.0475
    employee = 0 if gross_salary > limit else ceil(employee_rate * gross_salary * days_worked / days_in_month)
    employer = 0 if gross_salary > limit else ceil(employer_rate * gross_salary * days_worked / days_in_month)
    return employee, employer, employee + employer


def pf_contribution(gross_salary: int, days_worked: int, days_in_month: int) -> tuple[int, int, int]:
    limit = 15000
    employee_rate = 0.12
    employer_rate = 0.12 + 0.011 + 0.005 + 0.0001
    employee = 0 if gross_salary > limit else ceil(employee_rate * gross_salary * days_worked / days_in_month)
    employer = 0 if gross_salary > limit else ceil(employer_rate * gross_salary * days_worked / days_in_month)
    return employee, employer, employee + employer
