from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_, select
from sqlalchemy.orm import Session

import brewman.schemas.daybook as schemas

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Daybook)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["daybook"]),
) -> schemas.Daybook:
    return schemas.Daybook(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        body=[],
    )


@router.get("/{start}/{finish}", response_model=schemas.Daybook)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["daybook"]),
) -> schemas.Daybook:
    start_date = datetime.strptime(start, "%d-%b-%Y")
    finish_date = datetime.strptime(finish, "%d-%b-%Y")
    with SessionFuture() as db:
        body = build_report(start_date, finish_date, db)
        set_period(start, finish, request.session)
        return schemas.Daybook(start_date=start_date, finish_date=finish_date, body=body)


def build_report(start_date: date, finish_date: date, db: Session) -> list[schemas.DaybookItem]:
    body = []

    query = (
        db.execute(
            select(Voucher)
            .join(Voucher.journals)
            .join(Journal.account)
            .where(
                Voucher.date_ >= start_date,
                Voucher.date_ <= finish_date,
                not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            )
            .order_by(Voucher.date_, Voucher.last_edit_date)
        )
        .unique()
        .scalars()
        .all()
    )

    for voucher in query:
        debit = Decimal(0)
        credit = Decimal(0)
        name_debit = ""
        name_credit = ""
        for journal in voucher.journals:
            if journal.debit == 1:
                debit += journal.amount
                name_debit += f"{journal.account.name} / "
            else:
                credit += journal.amount
                name_credit += f"{journal.account.name} / "
        name_debit = name_debit[:-3]
        name_credit = name_credit[:-3]

        body.append(
            schemas.DaybookItem(
                id_=voucher.id,
                date_=voucher.date_,
                url=[
                    "/",
                    voucher.voucher_type.name.replace("_", "-").lower(),
                    str(voucher.id),
                ],
                type_=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                posted=voucher.posted,
                debit_text=name_debit,
                debit_amount=debit,
                credit_text=name_credit,
                credit_amount=credit,
            )
        )
    return body
