from fastapi import APIRouter, Security
from sqlalchemy import select
from sqlalchemy.orm import Session, contains_eager, joinedload

import brewman.schemas.non_contract_purchase as schemas

from brewman.core.security import get_current_active_user as get_user
from brewman.db.session import SessionFuture
from brewman.models.batch import Batch
from brewman.models.inventory import Inventory
from brewman.models.journal import Journal
from brewman.models.product import Product
from brewman.models.rate_contract import RateContract
from brewman.models.stock_keeping_unit import StockKeepingUnit
from brewman.models.voucher import Voucher
from brewman.models.voucher_type import VoucherType
from brewman.schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=list[schemas.NonContractPurchase])
def non_contract_purchases(
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> list[schemas.NonContractPurchase]:
    with SessionFuture() as db:
        return report(db)


def report(db: Session) -> list[schemas.NonContractPurchase]:
    rcs: list[RateContract] = (
        db.execute(  # type: ignore
            select(RateContract)
            .join(RateContract.vendor)
            .join(RateContract.items)
            .options(
                joinedload(RateContract.items, innerjoin=True),
                contains_eager(RateContract.items),
                joinedload(RateContract.vendor, innerjoin=True),
                contains_eager(RateContract.vendor),
            )
        )
        .unique()
        .scalars()
        .all()
    )
    list_: list[schemas.NonContractPurchase] = []
    for rc in rcs:
        for item in rc.items:
            invs = db.execute(
                select(Voucher.id, Voucher.date_, Inventory.id, Inventory.rate)
                .join(Inventory.batch)
                .join(Inventory.voucher)
                .join(Voucher.journals)
                .where(
                    Batch.sku_id == item.sku_id,
                    Voucher.date_ >= rc.valid_from,
                    Voucher.date_ <= rc.valid_till,
                    Voucher.voucher_type == VoucherType.PURCHASE,
                    Journal.account_id == rc.vendor_id,
                    Inventory.rate != item.price,
                )
            ).all()
            for inv in invs:
                p, u = db.execute(
                    select(Product.name, StockKeepingUnit.units)
                    .join(Product.skus)
                    .where(StockKeepingUnit.id == item.sku_id)
                ).one()
                list_.append(
                    schemas.NonContractPurchase(
                        date_=inv.date_,
                        vendor=rc.vendor.name,
                        product=f"{p} ({u})",
                        url=[
                            "/",
                            "purchase",
                            str(inv.id),
                        ],
                        contract_price=item.price,
                        purchase_price=inv.rate,
                    )
                )
    return list_
