import uuid

from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, Request, Security
from sqlalchemy.orm import Session, joinedload
from sqlalchemy.sql.expression import func, select

import brewman.schemas.product_ledger as schemas

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.ProductLedger)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> schemas.ProductLedger:
    return schemas.ProductLedger(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        product=None,
        body=[],
    )


@router.get("/{id_}", response_model=schemas.ProductLedger)
def show_data(
    id_: uuid.UUID,
    request: Request,
    s: str | None = None,
    f: str | None = None,
    user: UserToken = Security(get_user, scopes=["product-ledger"]),
) -> schemas.ProductLedger:
    with SessionFuture() as db:
        product = db.execute(select(Product).where(Product.id == id_)).scalar_one()
        start_date = s if s is not None else get_start_date(request.session)
        finish_date = f if f is not None else get_finish_date(request.session)
        body = build_report(
            product.id,
            datetime.strptime(start_date, "%d-%b-%Y"),
            datetime.strptime(finish_date, "%d-%b-%Y"),
            db,
        )
        set_period(start_date, finish_date, request.session)
        return schemas.ProductLedger(
            start_date=start_date,
            finish_date=finish_date,
            product=schemas.ProductLink(id_=product.id, name=product.name),
            body=body,
        )


def build_report(
    product_id: uuid.UUID, start_date: date, finish_date: date, db: Session
) -> list[schemas.ProductLedgerItem]:
    running_total_q, running_total_a, opening = opening_balance(product_id, start_date, db)
    body = opening

    query = db.execute(
        select(Voucher, Inventory, Journal, StockKeepingUnit)
        .join(Voucher.journals)
        .join(Voucher.inventories)
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .options(
            joinedload(Journal.account, innerjoin=True),
            joinedload(Journal.cost_centre, innerjoin=True),
        )
        .where(
            StockKeepingUnit.product_id == product_id,
            Journal.cost_centre_id != CostCentre.cost_centre_purchase(),
            Voucher.date_ >= start_date,
            Voucher.date_ <= finish_date,
        )
        .order_by(Voucher.date_, Voucher.last_edit_date)
    ).all()

    for voucher, inventory, journal, stockKeepingUnit in query:
        journal_debit = journal.debit * -1
        name = (
            journal.cost_centre.name
            if voucher.voucher_type in [VoucherType.ISSUE, VoucherType.CLOSING_STOCK]
            else journal.account.name
        )
        debit_q = inventory.quantity if journal_debit == 1 else None
        debit_a = inventory.amount if journal_debit == 1 else None
        debit_u = stockKeepingUnit.units if journal_debit == 1 else None
        credit_q = inventory.quantity if journal_debit != 1 else None
        credit_a = inventory.amount if journal_debit != 1 else None
        credit_u = stockKeepingUnit.units if journal_debit != 1 else None

        running_total_q += inventory.quantity * journal_debit
        running_total_a += inventory.amount * journal_debit

        body.append(
            schemas.ProductLedgerItem(
                id_=voucher.id,
                date_=voucher.date_,
                name=name,
                url=[
                    "/",
                    voucher.voucher_type.name.replace("_", "-").lower(),
                    str(voucher.id)
                    if voucher.voucher_type != VoucherType.CLOSING_STOCK
                    else voucher.date_.strftime("%d-%b-%Y"),
                ],
                type_=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                posted=voucher.posted or voucher.voucher_type in [VoucherType.ISSUE, VoucherType.CLOSING_STOCK],
                debit_quantity=debit_q,
                debit_amount=debit_a,
                debit_unit=debit_u,
                credit_quantity=credit_q,
                credit_amount=credit_a,
                credit_unit=credit_u,
                running_quantity=running_total_q,
                running_amount=running_total_a,
            )
        )

    return body


def opening_balance(
    product_id: uuid.UUID, start_date: date, db: Session
) -> tuple[Decimal, Decimal, list[schemas.ProductLedgerItem]]:
    row = db.execute(
        select(
            StockKeepingUnit.units,
            func.sum(Inventory.quantity * Journal.debit).label("quantity"),
            func.sum(Inventory.amount * Journal.debit).label("amount"),
        )
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Voucher.id == Inventory.voucher_id,
            Voucher.id == Journal.voucher_id,
            StockKeepingUnit.product_id == product_id,
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
            Voucher.date_ < start_date,
        )
        .group_by(StockKeepingUnit.units)
    ).all()

    quantity: Decimal = sum(r.quantity for r in row)
    amount: Decimal = sum(r.amount for r in row)

    return (
        quantity,
        amount,
        [
            schemas.ProductLedgerItem(
                id_=None,
                date_=start_date,
                name="Opening Balance",
                url=[],
                type_="Opening Balance",
                narration="",
                posted=True,
                debit_quantity=q,
                debit_amount=a,
                debit_unit=u,
                credit_quantity=Decimal(0),
                credit_amount=Decimal(0),
                credit_unit="",
                running_quantity=quantity,
                running_amount=amount,
            )
            for u, q, a in row
        ],
    )
